; This file is part of Intellidiscs.

; Intellidiscs is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; Intellidiscs is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with Intellidiscs.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2021 Phaser Cat Games LLC

; General utility routines

; Start a scene transition.
start_transition subroutine
    jsr     wait_for_vb             ; Sync the display.
    lda     #0                      ; Disable playfield and DMA.
    sta     sDMACTL
    sta     GRACTL                  ; Disable sprites.
    lda     #<ENDINT                ; Clear vertical blank interrupt.
    sta     VVBLKD
    lda     #>ENDINT
    sta     VVBLKD + 1
    jsr     STOPTUN                 ; Kill all sounds.
    jsr     CLEARTUN
    ; Fall through and wait for VBlank to update shadows.

; Busy-waits until vertical blank.
wait_for_vb subroutine
    lda     RCLKLO                  ; Read the low byte of the clock.
.loop
    cmp     RCLKLO                  ; Loop until it changes.
    beq     .loop
    rts

; Restore the graphics registers after a scene transition.
complete_transition subroutine
    jsr     wait_for_vb             ; Sync the display.
    mva     #$3e,sDMACTL            ; Reenable playfield and DMA.
    mva     #$03,GRACTL             ; Reenable sprites.
    jsr     STARTTUN
    jmp     wait_for_vb             ; Sync up and return.

; Clear the player/missile memory pages.
clear_sprites subroutine
    ldy     #0
    tya
.loop
    sta     $1300,y
    sta     $1400,y
    sta     $1500,y
    sta     $1600,y
    sta     $1700,y
    dey
    bne     .loop
    rts

; Pause for the number of frames in y.
wait_for_y_frames subroutine
    jsr     wait_for_vb
    dey
    bne     wait_for_y_frames
    rts

; Copies the memory located at $fc, $fd to $fe, $ff.
; Number of bytes to copy should be in y.
copy_memory subroutine
    dey
    cpy     #$ff
    beq     .done                   ; Done after underflow.
    lda     ($fc),y                 ; Read a byte.
    sta     ($fe),y                 ; Store a byte.
    jmp     copy_memory             ; Do it again.
.done
    rts

; Invert the value in the accumulator.
invert subroutine
    eor     #$ff
    clc
    adc     #1
    rts

; Puts the higher of the values in y and the
; accumulator into the accumulator.  Clobbers $ff.
get_max_value subroutine
    sty     $ff
    cmp     $ff
    bcs     .accumulator_wins
    tya
.accumulator_wins
    rts

; Same as get_max_value, but gets the minimum.
get_min_value subroutine
    sty     $ff
    cmp     $ff
    bcc     .accumulator_wins
    tya
.accumulator_wins
    rts

; Copy memory in reverse.  Same arguments as copy_memory.
copy_memory_reverse subroutine
    mva     #0,$fa                  ; Target index.
    sty     $fb                     ; Save the source index.
.loop
    ldy     $fb                     ; Get the source index.
    dey
    cpy     #$ff
    beq     .done                   ; Done after underflow.
    lda     ($fc),y                 ; Read a byte.
    ldy     $fa
    sta     ($fe),y                 ; Store byte.
    inc     $fa                     ; Increment the target index.
    dec     $fb                     ; Decrement the source index.
    jmp     .loop                   ; Do it again.
.done
    rts

; Move all players and missiles offscreen.
hide_all_sprites subroutine
    ldy     #7
    lda     #0
.loop
    sta     HPOSP0,y
    dey
    bpl     .loop
    rts

; Transfer text from the address in $fc-$fd to the
; address in $fe-$ff.  Copies until a 0 is found.
; This version does not handle lowercase letters,
; but is faster as a result.
copy_text subroutine
    sec
    ldy     #0
.loop
    lda     ($fc),y
    beq     .done
    sbc     #$20
    sta     ($fe),y
    iny
    jmp     .loop
.done
    rts

; As above, but this version handles lowercase letters.
copy_text_full subroutine
    ldy     #0
.loop
    lda     ($fc),y
    beq     .done
    cmp     #97
    bcs     .lowercase
    sec
    sbc     #$20
.lowercase
    sta     ($fe),y
    iny
    jmp     .loop
.done
    rts

; Decompress a monochromie image.  Expects the source
; address in $fc-$fd and the target address in $fe-$ff.
decompress_monochrome subroutine
    ldy     #8                      ; Bits remaining on this byte.
    sty     $f2
    ldy     #0
.run_loop
    lda     ($fc),y                 ; Get the run.
    beq     .done                   ; Run of 0 terminates.
    tax                             ; Save the run length.
    jsr     .next_read_byte
.process_value
    clc
    lda     #$ff                    ; Add the value to $ff
    adc     ($fc),y                 ; to set the carry bit.
    rol     $f3                     ; Rotate the bit in.
    dec     $f2                     ; Eight bits filled?
    bne     .continue_byte
    lda     $f3                     ; Get the finished byte.
    sta     ($fe),y                 ; Write the byte.
    inc     $fe                     ; Update the index.
    bne     .no_write_overflow
    inc     $ff                     ; Go to the next page.
.no_write_overflow
    lda     #8                      ; Reset the bit count.
    sta     $f2
.continue_byte
    dex                             ; Reduce the run.
    bne     .process_value
    jsr     .next_read_byte         ; Run over, go to next run.
    jmp     .run_loop
.next_read_byte
    inc     $fc
    bne     .done
    inc     $fd                     ; Next page.
.done
    rts

